// --------------------------------------------------------------------------
//                     Tile Studio Scroll Demo Example
//
// Mike Wiering (mike.wiering@cs.kun.nl)
// Tile Studio: http://www.cs.kun.nl/is/ts/
// --------------------------------------------------------------------------

#define SCREEN_WIDTH   640
#define SCREEN_HEIGHT  480
#define BITS_PER_PIXEL  16

#define BACK_BUFFERS  1

#define NAME    "Scroll"
#define TITLE   "Tile Studio Scroll Demo"

#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif

#define DIRECTDRAW_VERSION 0x0700

#include <stdio.h>
#include <stdarg.h>
#include <windows.h>
#include <ddraw.h>

#include "resource.h"
#include "ddutil.h"

#include "gfx\ScrollDemoGfx.c"


// --------------------------------------------------------------------------
// Global Variables
// --------------------------------------------------------------------------

HINSTANCE               g_hInstance;
LPDIRECTDRAW7           g_pDD         = NULL;   // DirectDraw object
LPDIRECTDRAWSURFACE7    g_pDDSPrimary = NULL;   // DirectDraw primary surface
LPDIRECTDRAWSURFACE7    g_pDDSBack    = NULL;   // DirectDraw back surface
RECT                    g_rcWindow;             // Saves the window size & pos.
RECT                    g_rcViewport;           // Pos. & size to blt from
RECT                    g_rcScreen;             // Screen pos. for blt
BOOL                    g_bActive     = FALSE;  // App is running/active
BOOL                    g_bReady      = FALSE;  // App is ready for updates
BOOL                    g_bWindowed   = FALSE;  // App is in windowed mode

static int ScreenWidth  = SCREEN_WIDTH;
static int ScreenHeight = SCREEN_HEIGHT;
static int ColorDepth   = BITS_PER_PIXEL;

static int LayerCount = 0;

static int XView = 0;
static int YView = 0;
static int XScrollSpeed = 0;
static int YScrollSpeed = 0;

static int iMaxXScrollSpeed = 8;
static int iMaxYScrollSpeed = 8;
static int iActualXView = 0;
static int iActualYView = 0;

#define MAX_VALUE 0x7FFFFFFF


static inline int intdiv (int x, int y)
{
  return (x < 0)? (x - y + 1) / y : x / y;
}

static inline int Clip (int x, int min, int max)
{
  return (x < min)? min : ((x > max)? max : x);
}

static inline int Min (int x, int y)
{
  return (x < y)? x : y;
}

static inline int Max (int x, int y)
{
  return (x > y)? x : y;
}

int FreeBitmaps ();


// --------------------------------------------------------------------------
// Name: ReleaseAllObjects()
// Desc: Release all DDraw objects we use
// --------------------------------------------------------------------------

HRESULT
ReleaseAllObjects (HWND hWnd)
{
    if (g_pDD != NULL)
    {
        g_pDD->SetCooperativeLevel (hWnd, DDSCL_NORMAL);
        if (g_pDDSBack != NULL)
        {
            g_pDDSBack->Release ();
            g_pDDSBack = NULL;
        }
        if (g_pDDSPrimary != NULL)
        {
            g_pDDSPrimary->Release ();
            g_pDDSPrimary = NULL;
        }
    }

    FreeBitmaps ();

    return DD_OK;
}


// --------------------------------------------------------------------------
// Name: InitFail()
// Desc: This function is called if an initialization function fails
// --------------------------------------------------------------------------

HRESULT
InitFail (HWND hWnd, HRESULT hRet, LPCTSTR szError, ...)
{
    char            szBuff[128];
    va_list         vl;

    va_start (vl, szError);
    vsprintf (szBuff, szError, vl);
    ReleaseAllObjects (hWnd);
    MessageBox (hWnd, szBuff, TITLE, MB_OK);
    DestroyWindow (hWnd);
    va_end (vl);
    return hRet;
}



#define MAX_BITMAP 32

typedef struct BITMAP_INFO
{
  LPDIRECTDRAWSURFACE7 lpDDSBitmap;
  int iBitmapWidth;
  int iBitmapHeight;
  int iBlockWidth;
  int iBlockHeight;
  int iBlockCountX;
  int iBlockCountY;
  BOOL bTransparent;
  int iSequences;
  int *ipSequenceLength;
  int *ipSequencePosition;
  int *ipSequenceCounter;
  int **ippSequenceData;
  DWORD *dwpFillColor;
} BITMAP_INFO;

static BITMAP_INFO biBitmap[MAX_BITMAP];


static int InitBitmap (const struct GFX_BITMAP_DATA * gfxBitmapData)
{
  HRESULT ret;
  DDSURFACEDESC2 ddsd;
  IDirectDrawSurface7 *lpDDS;

  BITMAP_INFO *bip = &biBitmap[0];
  int id = 0;
  while (bip->lpDDSBitmap)
  {
    bip++;
    id++;
    if (id >= MAX_BITMAP)
      return -1;
  }

  memset (&ddsd, 0, sizeof (ddsd));
  ddsd.dwSize = sizeof (ddsd);
  ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;

  ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;  // | DDSCAPS_SYSTEMMEMORY;
  ddsd.dwWidth = gfxBitmapData->BitmapWidth;
  ddsd.dwHeight = gfxBitmapData->BitmapHeight;
  ret = g_pDD->CreateSurface (&ddsd, &lpDDS, NULL);
  if (ret != DD_OK)
    return -1;
//     return InitFail (hWnd, hRet, "CreateSurface failed");

  ret = DDReLoadBitmap (lpDDS, gfxBitmapData->BitmapIdentifier);
  if (ret != DD_OK)
    return -1;

  bip->lpDDSBitmap = lpDDS;
  bip->bTransparent = FALSE;
  bip->iBitmapWidth = gfxBitmapData->BitmapWidth;
  bip->iBitmapHeight = gfxBitmapData->BitmapHeight;
  bip->iBlockWidth = gfxBitmapData->TileWidth;
  bip->iBlockHeight = gfxBitmapData->TileHeight;
  bip->iBlockCountX = gfxBitmapData->TileCountX;
  bip->iBlockCountY = gfxBitmapData->TileCountY;
  bip->iSequences = gfxBitmapData->Sequences;
  bip->ippSequenceData = gfxBitmapData->SequenceData;
  bip->ipSequenceLength = gfxBitmapData->SequenceLength;
  bip->ipSequencePosition = gfxBitmapData->SequencePosition;
  bip->ipSequenceCounter = gfxBitmapData->SequenceCounter;
  bip->dwpFillColor = NULL;

  return id;
}

COLORREF GetPix (BITMAP_INFO *bip, int x, int y)
{
  HDC hdc;
  COLORREF rgb;

  while (x < 0)
    x += bip->iBitmapWidth;
  while (y < 0)
    y += bip->iBitmapHeight;

  x %= bip->iBitmapWidth;
  y %= bip->iBitmapHeight;

  if (bip->lpDDSBitmap->GetDC (&hdc) != DD_OK)
    return -1;
  rgb = GetPixel (hdc, x, y);
  bip->lpDDSBitmap->ReleaseDC (hdc);
  return rgb;
}


int SetFillData (int id, int *points, DWORD *fillcolor, int count)
{
  BITMAP_INFO *bip = &biBitmap[id];
  COLORREF rgb;

  if (!bip->lpDDSBitmap)
    return -1;

  for (int i = 0; i < count; i++)
  {
    rgb = GetPix (bip, points[2 * i], points[2 * i + 1]);
    fillcolor[i] = DDColorMatch (g_pDDSBack, rgb);
  }

  bip->dwpFillColor = &fillcolor[0];
  return 0;
}

static int BitmapDone (int id)
{
  BITMAP_INFO *bip = &biBitmap[id];

  if (bip->lpDDSBitmap)
    bip->lpDDSBitmap->Release ();
  bip->lpDDSBitmap = NULL;
  return 0;
}

static int FreeBitmaps ()
{
  for (int id = 0; id < MAX_BITMAP; id++)
    BitmapDone (id);
  return 0;
}

int SetTransparentColor (int id, const struct GFX_BITMAP_DATA * gfxBitmapData)
{
  COLORREF rgb;
  BITMAP_INFO *bip = &biBitmap[id];
  int x = gfxBitmapData->TransparentX;
  int y = gfxBitmapData->TransparentY;

  if (!bip->lpDDSBitmap)
    return -1;

  rgb = GetPix (bip, x, y);

  DDSetColorKey (bip->lpDDSBitmap, rgb);
  bip->bTransparent = TRUE;

  return 0;
}

int RunSequences ()
{
  for (int id = 0; id < MAX_BITMAP; id++)
  {
    BITMAP_INFO *bip = &biBitmap[id];
    if (bip->lpDDSBitmap)
      for (int i = 0; i < bip->iSequences; i++)
        if (--bip->ipSequenceCounter[i] <= 0)
        {
          if (++bip->ipSequencePosition[i] >= bip->ipSequenceLength[i])
            bip->ipSequencePosition[i] = 0;
          bip->ipSequenceCounter[i] =
              bip->ippSequenceData[i][2 * bip->ipSequencePosition[i] + 1];
        }
  }
  return 0;
}



static unsigned short *sBoundMap = NULL;
static int BoundMapWidth = 0;
static int BoundMapHeight = 0;
static int BoundBlockWidth = 0;
static int BoundBlockHeight = 0;
static int TotalWidth = 0;
static int TotalHeight = 0;


static int InitBoundMap (unsigned short *map, int w, int h, int bid)
//                         int objectstartvalue, int startobjx, int startobjy)
{
  BITMAP_INFO *bip = &biBitmap[bid];

  sBoundMap = map;
  BoundMapWidth = w;
  BoundMapHeight = h;
  BoundBlockWidth = bip->iBlockWidth;
  BoundBlockHeight = bip->iBlockHeight;
  TotalWidth = bip->iBlockWidth * w;
  TotalHeight = bip->iBlockHeight * h;

//  iObjectValueStart = objectstartvalue;
//  iStartObjX = startobjx;
//  iStartObjY = startobjy;

  XView = 0;
  YView = 0;

  return 0;
}

/* move XView and YView towards iActualXView and iActualYView */
int CorrectView ()
{
  if (iActualXView > XView)
    XView = Min (XView + iMaxXScrollSpeed, iActualXView);
  if (iActualYView > YView)
    YView = Min (YView + iMaxYScrollSpeed, iActualYView);
  if (iActualXView < XView)
    XView = Max (XView - iMaxXScrollSpeed, iActualXView);
  if (iActualYView < YView)
    YView = Max (YView - iMaxYScrollSpeed, iActualYView);

  if (XView > TotalWidth - ScreenWidth)
  {
    XView = TotalWidth - ScreenWidth;
    iActualXView = XView;
    XScrollSpeed = 0;
  }
  if (YView > TotalHeight - ScreenHeight)
  {
    YView = TotalHeight - ScreenHeight;
    iActualYView = YView;
    YScrollSpeed = 0;
  }

  if (XView < 0)
  {
    XView = 0;
    iActualXView = XView;
    XScrollSpeed = 0;
  }
  if (YView < 0)
  {
    YView = 0;
    iActualYView = YView;
    YScrollSpeed = 0;
  }
  return 0;
}

static int MoveScreenTo (int x, int y)
{
  XView = x;
  YView = y;
  iActualXView = XView;
  iActualYView = YView;

  CorrectView ();
  return 0;
}

#define MAX_LAYER 8

typedef struct LAYERMAPINFO
{
  int iBitmapID;
  int iMapWidth;
  int iMapHeight;
  signed short *sMap;
  int iScrollFunction;
  int iHShift;
  int iVShift;
} LAYERMAPINFO;

static LAYERMAPINFO lmiLayerMapInfo[MAX_LAYER];


static int InitLayerMap (int bitmapid,
                         int mapwidth, int mapheight,
                         signed short *map,
                         int scrollfunction,
                         int hshift,
                         int vshift)
{
  if (LayerCount < MAX_LAYER)
  {
    LAYERMAPINFO *lmip = &lmiLayerMapInfo[LayerCount];
    lmip->iBitmapID = bitmapid;
    lmip->sMap = map;
    lmip->iMapWidth = mapwidth;
    lmip->iMapHeight = mapheight;
    lmip->iScrollFunction = scrollfunction;
    lmip->iHShift = hshift;
    lmip->iVShift = vshift;
    LayerCount++;
    return 0;
  }
  else
    return -1;
}


int FillRect (RECT r, COLORREF c)
{
  DDBLTFX ddbltfx;

  memset (&ddbltfx, 0, sizeof (ddbltfx));
  ddbltfx.dwSize = sizeof (ddbltfx);
  ddbltfx.dwFillColor = DDColorMatch (g_pDDSBack, c);
  g_pDDSBack->Blt (&r, NULL, NULL, DDBLT_COLORFILL | DDBLT_WAIT, &ddbltfx);
  return 0;
}

int CalcLayerPos (int scrollfunction, int *x, int *y)
{
  switch (scrollfunction)
  {
    case 0:
    {
      *x = XView;
      *y = YView;
    }
    break;
    default:
    {
      *x = XView / scrollfunction;
      *y = YView / scrollfunction;
    }
  }
  return 0;
}

int DrawLayer (int layer)
{
  int front = 0;
  RECT r, src, dst;
  int xstart, ystart;
  int x, y;
  int w, h;
  int mapx, mapy;
  int mapw, maph;
  int curw, curh;
  int mappos, mapblock;
  int bmpx, bmpy;
  int fx;
  int maxblock;
  BOOL fill;
  DWORD fillcolor;
  BOOL mir, ups;
  LAYERMAPINFO *lmip = &lmiLayerMapInfo[layer];
  DDBLTFX ddbltfx;
  int bltflags, bltfastflags;
  BOOL fast;
  int frontblocks = 0;

  if ((layer < LayerCount) && (lmip))
  {
    BITMAP_INFO *bip = &biBitmap[lmip->iBitmapID];

    if (bip)
    {
      w = bip->iBlockWidth;
      h = bip->iBlockHeight;

      maxblock = bip->iBlockCountX * bip->iBlockCountY;

      mapw = lmip->iMapWidth;
      maph = lmip->iMapHeight;

      // calculate layer position
      CalcLayerPos (lmip->iScrollFunction, &xstart, &ystart);
      xstart += lmip->iHShift;
      ystart += lmip->iVShift;
      while (xstart < 0)
        xstart += mapw * w;
      while (ystart < 0)
        ystart += maph * h;

      mapy = ystart / h;
      r.top = ystart % h;

      for (y = 0; y < ScreenHeight; )
      {
        mapx = xstart / w;
        r.left = xstart % w;
        curh = Min (ScreenHeight - y, h - r.top);

        for (x = 0; x < ScreenWidth; )
        {
          curw = Min (ScreenWidth - x, w - r.left);

          // calculate position in map, read block number
          mappos = ((mapx % mapw) + ((mapy % maph) * mapw));
          mapblock = *(signed short *) &lmip->sMap[mappos];

          fill = FALSE;
          if (mapblock < 0)  // sequence ID
          {
            int seq = -mapblock - 1;
            if (seq < bip->iSequences)
            {
              int pos = bip->ipSequencePosition[seq];
              int *seqdata = bip->ippSequenceData[seq];
              mapblock = seqdata[2 * pos];
            }
            else
            {
              seq -= bip->iSequences;
              if (bip->dwpFillColor)
              {
                fillcolor = bip->dwpFillColor[seq];
                fill = TRUE;
              }
              mapblock = 0;
            }
          }

          if (front)
          {
            if (mapblock >= 0x4000)
              mapblock -= (0x4000 - 1);
            else
              mapblock = 0;
          }
          else
            if (mapblock >= 0x4000)
            {
              mapblock -= 0x4000;
              frontblocks++;
            }

          mapblock--;
          fx = (mapblock / maxblock) & 3;

          dst.left = x;
          dst.top = y;
          dst.right = x + curw;
          dst.bottom = y + curh;

          if (fill)
          {
            memset (&ddbltfx, 0, sizeof (ddbltfx));
            ddbltfx.dwSize = sizeof (ddbltfx);
            ddbltfx.dwFillColor = fillcolor;
            g_pDDSBack->Blt (&dst, NULL, NULL,
                             DDBLT_COLORFILL | DDBLT_WAIT, &ddbltfx);
          }

          if (mapblock >= 0)
          {
	    mapblock %= maxblock;
            bmpx = (mapblock % bip->iBlockCountX) * w;
            bmpy = (mapblock / bip->iBlockCountX) * h;
            src.left = r.left + bmpx;
            src.top = r.top + bmpy;

            mir = FALSE;
            ups = FALSE;
            if ((fx & 1) == 1)
            {
              mir = TRUE;
              src.left = bmpx + w - curw - (r.left % w);
            }
            if ((fx & 2) == 2)
            {
              ups = TRUE;
              src.top = bmpy + h - curh - (r.top % w);
            }
            src.right = src.left + curw;
            src.bottom = src.top + curh;

            fast = TRUE;

            bltfastflags = DDBLTFAST_WAIT;
            bltflags = DDBLT_DDFX | DDBLT_WAIT;

            if (bip->bTransparent)
            {
              bltflags |= DDBLT_KEYSRC;
              bltfastflags |= DDBLTFAST_SRCCOLORKEY;
            }

            //  ddbltfx.dwDDFX = 0;
            memset (&ddbltfx, 0, sizeof (ddbltfx));
            ddbltfx.dwSize = sizeof (ddbltfx);

            if (mir)
            {
              ddbltfx.dwDDFX |= DDBLTFX_MIRRORLEFTRIGHT;
              fast = FALSE;
            }
            if (ups)
            {
              ddbltfx.dwDDFX |= DDBLTFX_MIRRORUPDOWN;
              fast = FALSE;
            }

            if (fast)
              g_pDDSBack->BltFast (dst.left,
                                   dst.top,
                                   bip->lpDDSBitmap,
                                   &src,
                                   bltfastflags);
            else
              g_pDDSBack->Blt (&dst,
                               bip->lpDDSBitmap,
                               &src,
                               bltflags,
                               &ddbltfx);
          }
          x += curw;
          r.left = 0;
          mapx++;
        }

        y += curh;
        r.top = 0;
        mapy++;
      }
    }
  }
  if (frontblocks > 0)
    return 1;
  else
    return 0;
}



int InitBitmaps ()
{
  int id = 0;
  memset (&biBitmap, 0, sizeof (biBitmap));
  LayerCount = 0;

  id = InitBitmap (&GfxBackGrBitmap);
  /* not transparent */
  InitLayerMap (id, GfxBackGrMap1MapWidth, GfxBackGrMap1MapHeight,
                &BackGrMap1MapData[0][0], 3, 0, 0);

  id = InitBitmap (&GfxBlocksBitmap);
  SetTransparentColor (id, &GfxBlocksBitmap);
  InitLayerMap (id, GfxBlocksLevel1MapWidth, GfxBlocksLevel1MapHeight,
                &BlocksLevel1MapData[0][0], 1, 0, 0);
  InitBoundMap (&BlocksLevel1BoundMapData[0][0],
                GfxBlocksLevel1MapWidth, GfxBlocksLevel1MapHeight, id);
  return 0;
}



// --------------------------------------------------------------------------
// UpdateFrame
// Draws the next frame
// --------------------------------------------------------------------------

int initialized = 0;

BOOL
UpdateFrame (HWND hWnd)
{
  int x = XView / BoundBlockWidth;
  int y = YView / BoundBlockHeight;
  int w = ScreenWidth / BoundBlockWidth;
  int h = ScreenHeight / BoundBlockHeight;

  if (!initialized)
  {
    XScrollSpeed = 0;
    YScrollSpeed = 0;
    CorrectView ();

    initialized++;
  }

  DrawLayer (0);
  DrawLayer (1);

  RunSequences ();

  iActualXView += XScrollSpeed;
  iActualYView += YScrollSpeed;
  CorrectView ();

  return TRUE;
}


// --------------------------------------------------------------------------
// Name: InitSurfaces()
// Desc: Create all the needed DDraw surfaces and set the coop level
// --------------------------------------------------------------------------

static HRESULT
InitSurfaces(HWND hWnd)
{
    HRESULT             hRet;
    DDSURFACEDESC2      ddsd;
    DDSCAPS2            ddscaps;
    LPDIRECTDRAWCLIPPER pClipper;

    if (g_bWindowed)
    {
        // Get normal windowed mode
        hRet = g_pDD->SetCooperativeLevel(hWnd, DDSCL_NORMAL);
        if (hRet != DD_OK)
            return InitFail(hWnd, hRet, "SetCooperativeLevel FAILED");

        // Get the dimensions of the viewport and screen bounds
        GetClientRect(hWnd, &g_rcViewport);
        GetClientRect(hWnd, &g_rcScreen);
        ClientToScreen(hWnd, (POINT*)&g_rcScreen.left);
        ClientToScreen(hWnd, (POINT*)&g_rcScreen.right);

        // Create the primary surface
        ZeroMemory (&ddsd,sizeof (ddsd));
        ddsd.dwSize = sizeof (ddsd);
        ddsd.dwFlags = DDSD_CAPS;
        ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
        hRet = g_pDD->CreateSurface (&ddsd, &g_pDDSPrimary, NULL);
        if (hRet != DD_OK)
            return InitFail (hWnd, hRet, "CreateSurface FAILED (1)");

        // Create a clipper object since this is for a Windowed render
        hRet = g_pDD->CreateClipper(0, &pClipper, NULL);
        if (hRet != DD_OK)
            return InitFail(hWnd, hRet, "CreateClipper FAILED");

        // Associate the clipper with the window
        pClipper->SetHWnd(0, hWnd);
        g_pDDSPrimary->SetClipper(pClipper);
        pClipper->Release();
        pClipper = NULL;

        // Get the backbuffer. For fullscreen mode, the backbuffer was created
        // along with the primary, but windowed mode still needs to create one.
        ddsd.dwFlags        = DDSD_WIDTH | DDSD_HEIGHT | DDSD_CAPS;
        ddsd.dwWidth        = ScreenWidth;
        ddsd.dwHeight       = ScreenHeight;
        ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
        hRet = g_pDD->CreateSurface (&ddsd, &g_pDDSBack, NULL);
        if (hRet != DD_OK)
            return InitFail (hWnd, hRet, "CreateSurface FAILED (2)");
    }
    else
    {
        // Get exclusive mode
        hRet = g_pDD->SetCooperativeLevel(hWnd, DDSCL_EXCLUSIVE |
                                                DDSCL_FULLSCREEN);
        if (hRet != DD_OK)
            return InitFail(hWnd, hRet, "SetCooperativeLevel FAILED");

        // Set the video mode to 640x480
        hRet = g_pDD->SetDisplayMode (ScreenWidth, ScreenHeight, ColorDepth, 0, 0);
        if (hRet != DD_OK)
            return InitFail(hWnd, hRet, "SetDisplayMode FAILED");

        // Get the dimensions of the viewport and screen bounds
        // Store the rectangle which contains the renderer
        SetRect (&g_rcViewport, 0, 0, ScreenWidth, ScreenHeight);
        memcpy (&g_rcScreen, &g_rcViewport, sizeof (RECT));

        // Create the primary surface with 1 back buffer
        ZeroMemory(&ddsd, sizeof(ddsd));
        ddsd.dwSize = sizeof(ddsd);
        ddsd.dwFlags = DDSD_CAPS |
                       DDSD_BACKBUFFERCOUNT;
        ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE |
                              DDSCAPS_FLIP |
                              DDSCAPS_COMPLEX;
        ddsd.dwBackBufferCount = BACK_BUFFERS;
        hRet = g_pDD->CreateSurface (&ddsd, &g_pDDSPrimary, NULL);
        if (hRet != DD_OK)
            return InitFail(hWnd, hRet, "CreateSurface FAILED (3)");

        ZeroMemory(&ddscaps, sizeof(ddscaps));
        ddscaps.dwCaps = DDSCAPS_BACKBUFFER;
        hRet = g_pDDSPrimary->GetAttachedSurface(&ddscaps, &g_pDDSBack);
        if (hRet != DD_OK)
            return InitFail(hWnd, hRet, "GetAttachedSurface FAILED");
    }

    InitBitmaps ();

    return DD_OK;
}




// --------------------------------------------------------------------------
// Name: ChangeCoopLevel()
// Desc: Called when the user wants to toggle between Full-Screen & Windowed
// --------------------------------------------------------------------------
HRESULT
ChangeCoopLevel(HWND hWnd )
{
    HRESULT hRet;

    // Release all objects that need to be re-created for the new device
    if (FAILED(hRet = ReleaseAllObjects(hWnd)))
        return InitFail(hWnd, hRet, "ReleaseAllObjects FAILED");

    // In case we're coming from a fullscreen mode, restore the window size
    if (g_bWindowed)
    {
        SetWindowPos(hWnd, HWND_NOTOPMOST, g_rcWindow.left, g_rcWindow.top,
                     (g_rcWindow.right - g_rcWindow.left),
                     (g_rcWindow.bottom - g_rcWindow.top), SWP_SHOWWINDOW);
    }

    // Re-create the surfaces
    hRet = InitSurfaces(hWnd);
    return hRet;
}



// --------------------------------------------------------------------------
// Name: AboutDlgProc()
// Desc: The About Dialog Box Procedure
// --------------------------------------------------------------------------
LRESULT CALLBACK
AboutDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    switch (uMsg)
    {
        case WM_INITDIALOG:
            return TRUE;
        case WM_COMMAND:
            switch (wParam)
            {
                case IDOK:
                    EndDialog(hDlg, TRUE);
                    return TRUE;
            }
            break;
    }

    return FALSE;
}




// --------------------------------------------------------------------------
// Name: MainWndProc()
// Desc: The Main Window Procedure
// --------------------------------------------------------------------------

LRESULT CALLBACK
MainWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    HRESULT         hRet;
    MINMAXINFO      *pMinMax;

    switch (msg)
    {
        case WM_ACTIVATE:
            // Pause if minimized
            g_bActive = !((BOOL)HIWORD(wParam));
            return 0L;

        case WM_COMMAND:
            // Handle all menu and accelerator commands
            switch (LOWORD(wParam))
            {
                case IDM_ABOUT:
                    DialogBox(g_hInstance, "AboutDlg", hWnd, (DLGPROC)AboutDlgProc);
                    return 0L;

                case IDM_TOGGLEFULLSCREEN:
                    // Toggle the fullscreen/window mode
                    if (g_bActive && g_bReady)
                    {
                        g_bReady = FALSE;
                        if (g_bWindowed)
                            GetWindowRect(hWnd, &g_rcWindow);
                        g_bWindowed = !g_bWindowed;
                        ChangeCoopLevel(hWnd);
                        g_bReady = TRUE;
                    }
                    return 0L;

                case IDM_EXIT:
                    // Received key/menu command to exit app
                    PostMessage(hWnd, WM_CLOSE, 0, 0);
                    return 0L;
            }
            break;

        case WM_DESTROY:
            // Clean up and close the app
            ReleaseAllObjects(hWnd);
            PostQuitMessage(0);
            return 0L;

        case WM_GETMINMAXINFO:
            // Fix the size of the window to 640x480 (client size)
            pMinMax = (MINMAXINFO *)lParam;
            pMinMax->ptMinTrackSize.x = ScreenWidth  + GetSystemMetrics (SM_CXSIZEFRAME) * 2;
            pMinMax->ptMinTrackSize.y = ScreenHeight + GetSystemMetrics (SM_CYSIZEFRAME) * 2
                                                     + GetSystemMetrics (SM_CYMENU);
            pMinMax->ptMaxTrackSize.x = pMinMax->ptMinTrackSize.x;
            pMinMax->ptMaxTrackSize.y = pMinMax->ptMinTrackSize.y;
            break;

        case WM_KEYDOWN:
            // Handle any non-accelerated key commands
            switch (wParam)
            {
                case VK_ESCAPE:
                    PostMessage(hWnd, WM_CLOSE, 0, 0);
                    return 0L;


                case VK_LEFT:
                    XScrollSpeed--;
                    return 0L;
                case VK_RIGHT:
                    XScrollSpeed++;
                    return 0L;
                case VK_UP:
                    YScrollSpeed--;
                    return 0L;
                case VK_DOWN:
                    YScrollSpeed++;
                    return 0L;

            }
            break;

        case WM_MOVE:
            // Retrieve the window position after a move
            if (g_bActive && g_bReady && g_bWindowed)
            {
                GetWindowRect(hWnd, &g_rcWindow);
                GetClientRect(hWnd, &g_rcViewport);
                GetClientRect(hWnd, &g_rcScreen);
                ClientToScreen(hWnd, (POINT*)&g_rcScreen.left);
                ClientToScreen(hWnd, (POINT*)&g_rcScreen.right);
            }
            break;

        case WM_PAINT:
            // Update the screen if we need to refresh
            if (g_bWindowed && g_bReady)
            {
                while (TRUE)
                {
                    // If we are in windowed mode, perform a blt.
                    hRet = g_pDDSPrimary->Blt(&g_rcScreen, g_pDDSBack,
                                              &g_rcViewport, DDBLT_WAIT,
                                              NULL);
                    if (hRet == DD_OK)
                        break;
                    if (hRet == DDERR_SURFACELOST)
                    {
                        hRet = g_pDDSPrimary->Restore();
                        if (hRet != DD_OK )
                            break;
                    }
                    if (hRet != DDERR_WASSTILLDRAWING)
                        break;
                }
            }
            break;

        case WM_SETCURSOR:
            // Display the cursor in the window if windowed
            if (g_bActive && g_bReady && !g_bWindowed)
            {
                SetCursor(NULL);
                return TRUE;
            }
            break;

        case WM_SIZE:
            // Check to see if we are losing our window...
            if (SIZE_MAXHIDE==wParam || SIZE_MINIMIZED==wParam)
                g_bActive = FALSE;
            else
                g_bActive = TRUE;
            break;
    }
    return DefWindowProc(hWnd, msg, wParam, lParam);
}


// --------------------------------------------------------------------------
// Name: WinMain()
// Desc: Entry point to the program. Initializes everything and calls
//       UpdateFrame() when idle from the message pump.
// --------------------------------------------------------------------------
int PASCAL
WinMain(HINSTANCE   hInstance,
        HINSTANCE   hPrevInstance,
        LPSTR       lpszCmdLine,
        int         nCmdShow)
{
    WNDCLASS        wc;
    MSG             msg;
    HWND            hWnd;
    HACCEL          hAccel;
    HRESULT         hRet;
    BOOL            page = FALSE;   // Which screen to render (Front or Back text)
    int             cx,cy;

    if (!hPrevInstance)
    {
        // Register the Window Class
        wc.lpszClassName = NAME;
        wc.lpfnWndProc = MainWndProc;
        wc.style = CS_VREDRAW | CS_HREDRAW;
        wc.hInstance = hInstance;
        wc.hIcon = LoadIcon (hInstance, MAKEINTRESOURCE (IDI_MAIN_ICON));
        wc.hCursor = LoadCursor (NULL, IDC_ARROW);
        wc.hbrBackground = (HBRUSH) (COLOR_WINDOW + 1);
        wc.lpszMenuName = MAKEINTRESOURCE (IDR_MENU);
        wc.cbClsExtra = 0;
        wc.cbWndExtra = 0;
        RegisterClass (&wc);
    }

    g_hInstance = hInstance;

    // Calculate the proper size for the window given a client of 640x480
    cx = ScreenWidth  + GetSystemMetrics (SM_CXSIZEFRAME) * 2;
    cy = ScreenHeight + GetSystemMetrics (SM_CYSIZEFRAME) * 2 +
                        GetSystemMetrics (SM_CYMENU);
    // Create and Show the Main Window
    hWnd = CreateWindowEx(0,
                          NAME,
                          TITLE,
                          WS_OVERLAPPEDWINDOW,
                          CW_USEDEFAULT,
                          CW_USEDEFAULT,
                          cx,
                          cy,
                          NULL,
                          NULL,
                          hInstance,
                          NULL);
    if (hWnd == NULL)
        return FALSE;
    ShowWindow (hWnd, nCmdShow);
    UpdateWindow (hWnd);

    // Save the window size/pos for switching modes
    GetWindowRect(hWnd, &g_rcWindow);

    // Load keyboard accelerators
    hAccel = LoadAccelerators (g_hInstance, MAKEINTRESOURCE (IDR_MAIN_ACCEL));

    // Create the main DirectDraw object
    hRet = DirectDrawCreateEx (NULL, (VOID**) &g_pDD, IID_IDirectDraw7, NULL);
    if (FAILED (hRet))
        return InitFail(hWnd, hRet, "DirectDrawCreateEx FAILED");

    // Initialize all the surfaces we need
    hRet = InitSurfaces (hWnd);
    if (FAILED (hRet))
        return FALSE;

    g_bReady = TRUE;

    // ----------------------------------------------------------------------
    //                          The message loop
    // ----------------------------------------------------------------------
    while (TRUE)
    {
        if (PeekMessage (&msg, NULL, 0, 0, PM_NOREMOVE))
        {
            if (!GetMessage (&msg, NULL, 0, 0))
                break;
            // Translate and dispatch the message
            if (0 == TranslateAccelerator (hWnd, hAccel, &msg))
            {
                TranslateMessage (&msg);
                DispatchMessage (&msg);
            }
        }
        else
            if (g_bActive && g_bReady)
            {
                // ----------------------------------------------------------
                //                    Idle processing
                // ----------------------------------------------------------

                // Update the background and flip every time the timer ticks
                UpdateFrame (hWnd);
                page = !page;
                while (TRUE)
                {
                    // If we are in windowed mode, perform a blt.
                    if (g_bWindowed)
                    {
                        hRet = g_pDDSPrimary->Blt (&g_rcScreen, g_pDDSBack,
                                                   &g_rcViewport, DDBLT_WAIT,
                                                   NULL);
                    }
                    else
                    {
                        // Else we are in fullscreen mode, so perform a flip.
                        hRet = g_pDDSPrimary->Flip (NULL, 0L);
                    }
                    if (hRet == DD_OK)
                        break;
                    if (hRet == DDERR_SURFACELOST)
                    {
                        hRet = g_pDDSPrimary->Restore ();
                        if (hRet != DD_OK)
                            break;
                    }
                    if (hRet != DDERR_WASSTILLDRAWING)
                        break;
                }
            }
            else
            {
                // Make sure we go to sleep if we have nothing else to do
                WaitMessage();
            }
    }
    // Release the main DDraw interface
    if (g_pDD)
        g_pDD->Release();

    return msg.wParam;
}

